# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
import socket
from collections import OrderedDict, namedtuple

import redis
import six
import sqlalchemy as sa
import sqlalchemy.dialects
import sqlalchemy.exc
from flask import current_app, flash, redirect, \
    render_template, request, session, url_for
from pathlib import Path
from six import text_type

from abilian.core.commands import config as cmd_config
from abilian.core.extensions import csrf, db
from abilian.core.models.subjects import User
from abilian.services import get_service
from abilian.services.security import Admin, Anonymous
from abilian.web.blueprints import Blueprint

logger = logging.getLogger(__name__)
setup = Blueprint(
    'setup',
    __name__,
    allowed_roles=Anonymous,
    template_folder='templates',
)

# list supported dialects and detect unavailable ones due to missing dbapi
# module ('psycopg2' missing for example)

_dialects = OrderedDict((
    ('sqlite', 'SQLite (for demo)'),
    ('postgresql', 'PostgreSQL'),
))

_dialects_unavailable = OrderedDict()

for dialect, _label in _dialects.items():
    d = sa.dialects.registry.load(dialect)
    try:
        d.dbapi()
    except ImportError as e:
        _dialects_unavailable[dialect] = str(e)

# enumerate steps for left column progress
Step = namedtuple('SetupStep', ('name', 'endpoint', 'title', 'description'))

_setup_steps = (
    Step('db', 'step_db', 'Setup Database', 'Setup basic database connection'),
    Step('redis', 'step_redis', 'Setup Redis', 'Redis connection'),
    Step(
        'site_info',
        'step_site_info',
        'Basic site informations',
        'Site name, admin email...',
    ),
    Step('admin_account', 'step_admin_account', 'Admin account', None),
    Step('finalize', 'finalize', 'Finalize', None),
)


@setup.before_request
def step_progress():
    session.permanent = False  # limit lifetime to browser session
    current_step = next_step = prev_step = None

    for step in _setup_steps:
        if current_step is not None:
            next_step = step
            break
        endpoint = '{}.{}'.format(setup.name, step.endpoint)
        if endpoint == request.endpoint:
            current_step = step
        else:
            prev_step = step

    request.setup_step_progress = {
        'current_step': current_step,
        'prev_step': prev_step,
        'next_step': next_step,
    }


@setup.context_processor
def common_context():
    ctx = {
        'setup_steps': _setup_steps,
        'dialects': _dialects,
        'dialects_unavailable': _dialects_unavailable,
        'validated_steps': session_get('validated', ()),
    }
    ctx.update(request.setup_step_progress)
    return ctx


# session helpers
def session_set(key, data):
    session.setdefault('abilian_setup', {})[key] = data


def session_get(key, default=None):
    return session.setdefault('abilian_setup', {}).get(key, default)


def session_clear():
    if 'abilian_setup' in session:
        del session['abilian_setup']


def step_validated(step_endpoint, valid=True):
    validated = set(session_get('validated', ()))
    if valid:
        validated.add(step_endpoint)
    else:
        if step_endpoint in validated:
            validated.remove(step_endpoint)

    session_set('validated', tuple(validated))


#
# DB Setup
#
@csrf.exempt
@setup.route('/', methods=['GET', 'POST'])
def step_db():
    if request.method == 'POST':
        return step_db_validate()
    return step_db_form()


def step_db_form():
    return render_template(
        'setupwizard/step_db.html',
        data=session_get('db', {}),
    )


def step_db_validate():
    form = request.form
    dialect = form['dialect']
    username = form.get('username', '').strip()
    password = form.get('password', '').strip()
    host = form.get('host', '').strip()
    port = form.get('port', '').strip()
    database = form.get('database', '').strip()

    # build db_uri
    db_uri = '{}://'.format(dialect)

    if dialect != 'sqlite':
        if username:
            db_uri += username
            # check password only if we have a username
            if password:
                db_uri += ':' + password

        # FIXME: it is an error to have a username and no host,
        # SA will interpret username:password as host:port
        if host:
            db_uri += '@' + host
            port = port
            if port:
                db_uri += ':' + port

    db_uri += '/'

    if dialect == 'sqlite' and (not database or database == ':memory:'):
        database = text_type(
            Path(current_app.instance_path) / 'data' / 'sqlite.db',
        )

    db_uri += database

    # store in session
    db_params = {
        'uri': db_uri,
        'dialect': dialect,
        'username': username,
        'password': password,
        'host': host,
        'port': port,
        'database': database,
    }
    session_set('db', db_params)

    # test connection
    engine = sa.create_engine(db_uri)
    error = None
    try:
        conn = engine.connect()
    except sa.exc.OperationalError as e:
        error = str(e)
    else:
        conn.close()
        del conn

    engine.dispose()

    # FIXME: at this point we have not tested the connection is valid, i.e, we
    # have CREATE, SELECT, UPDATE, DELETE rights on the database

    if error:
        flash(error, 'error')
        return step_db_form()

    step_validated('db')
    next_step = request.setup_step_progress['next_step']
    endpoint = '{}.{}'.format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Redis
#
@csrf.exempt
@setup.route('/redis', methods=['GET', 'POST'])
def step_redis():
    if request.method == 'POST':
        return step_redis_validate()
    return step_redis_form()


def step_redis_form():
    return render_template(
        'setupwizard/step_redis.html',
        data=session_get('redis', {}),
    )


def step_redis_validate():
    form = request.form
    data = dict(
        host=form.get('host', 'localhost').strip(),
        port=form.get('port', '').strip() or '6379',
        db=form.get('db', '').strip() or '1',
    )

    for k in ('port', 'db'):
        try:
            data[k] = int(data[k])
        except ValueError:
            pass

    data['uri'] = 'redis://{host}:{port}/{db}'.format(**data)
    session_set('redis', data)
    error = None

    try:
        r = redis.StrictRedis(
            host=data['host'],
            port=data['port'],
            db=data['db'],
        )
    except Exception as e:
        error = 'Connection error, check parameters'
        raise e

    try:
        r.info()
    except redis.exceptions.InvalidResponse:
        error = (
            "Connection error: doesn't look like it's a redis server. "
            "Verify host and port are those of your redis server."
        )
    except redis.exceptions.ResponseError as e:
        error = 'Redis server response: {}'.format(e)
    except redis.exceptions.RedisError as e:
        error = 'Unknown redis error ({})'.format(e)

    if error:
        flash(error, 'error')
        return step_redis_form()

    step_validated('redis')
    next_step = request.setup_step_progress['next_step']
    endpoint = '{}.{}'.format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Site info
#
@csrf.exempt
@setup.route('/site_info', methods=['GET', 'POST'])
def step_site_info():
    if request.method == 'POST':
        return step_site_info_validate()
    return step_site_info_form()


def get_possible_hostnames():
    hostname = socket.gethostname()
    fqdn = socket.getfqdn()
    names = {'localhost': ['127.0.0.1']}

    for name in (hostname, fqdn):
        try:
            name, aliases, ips = socket.gethostbyname_ex(name)
        except socket.error:
            continue
        names.setdefault(name, []).extend(ips)
        for a in aliases:
            names.setdefault(a, []).extend(ips)

    return sorted(
        '{} ({})'.format(name, ', '.join(sorted(set(ips))))
        for name, ips in six.iteritems(names)
    )


def step_site_info_form():
    cfg = current_app.config
    default_data = {
        'sitename': cfg.get('SITE_NAME', '') or '',
        'mailsender': cfg.get('MAIL_SENDER', '') or '',
    }
    return render_template(
        'setupwizard/step_site_info.html',
        data=session_get('site_info', default_data),
        suggested_hosts=get_possible_hostnames(),
    )


def step_site_info_validate():
    form = request.form
    data = {
        'sitename': form.get('sitename', '').strip(),
        'mailsender': form.get('mailsender', '').strip(),
        'server_mode': form.get('server_mode', '').strip(),
    }

    session_set('site_info', data)
    step_validated('site_info')
    next_step = request.setup_step_progress['next_step']
    endpoint = '{}.{}'.format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Admin account creation
#
@csrf.exempt
@setup.route('/admin_account', methods=['GET', 'POST'])
def step_admin_account():
    if request.method == 'POST':
        return step_admin_account_validate()
    return step_admin_account_form()


def step_admin_account_form():
    return render_template(
        'setupwizard/step_admin_account.html',
        data=session_get('admin_account', {}),
    )


def step_admin_account_validate():
    form = request.form
    password = form.get('password', '').strip()
    confirm_password = form.get('confirm_password', '').strip()
    admin_user = {
        'email': form.get('email', '').strip(),
        'name': form.get('name', '').strip(),
        'firstname': form.get('firstname', '').strip(),
        'password': password,
        'confirm_password': confirm_password,
    }
    session_set('admin_account', admin_user)

    if password != confirm_password:
        flash('Password fields don\'t match', 'error')
        return step_admin_account_form()

    step_validated('admin_account')
    next_step = request.setup_step_progress['next_step']
    endpoint = '{}.{}'.format(setup.name, next_step.endpoint)
    return redirect(url_for(endpoint))


#
# Finalize
#
@csrf.exempt
@setup.route('/finalize', methods=['GET', 'POST'])
def finalize():
    validated = session_get('validated')
    assert all(s.name in validated for s in _setup_steps[:-1])

    if request.method == 'POST':
        return finalize_validate()
    return finalize_form()


def finalize_form():
    file_location = Path(current_app.instance_path) / 'config.py'
    return render_template(
        'setupwizard/finalize.html',
        file_location=file_location,
    )


def finalize_validate():
    config_file = Path(current_app.instance_path) / 'config.py'
    logging_file = Path(current_app.instance_path) / 'logging.yml'

    assert not config_file.exists()
    config = cmd_config.DefaultConfig(logging_file='logging.yml')
    config.SQLALCHEMY_DATABASE_URI = session_get('db')['uri']

    redis_uri = session_get('redis')['uri']
    config.REDIS_URI = redis_uri
    config.BROKER_URL = redis_uri
    config.CELERY_RESULT_BACKEND = redis_uri

    d = session_get('site_info')
    config.SITE_NAME = d['sitename']
    config.MAIL_SENDER = d['mailsender']

    is_production = d['server_mode'] == 'production'
    config.PRODUCTION = is_production
    config.DEBUG = not is_production
    config.DEBUG_TB_ENABLED = config.DEBUG
    config.CELERY_ALWAYS_EAGER = not is_production

    cmd_config.write_config(config_file, config)
    cmd_config.maybe_write_logging(logging_file)

    admin_account = session_get('admin_account')
    # create a new app that will be configured with new config,
    # to create database and admin_user
    setup_app = current_app._get_current_object()
    app = setup_app.__class__(
        setup_app.import_name,
        static_url_path=setup_app.static_url_path,
        static_folder=setup_app.static_folder,
        template_folder=setup_app.template_folder,
        instance_path=setup_app.instance_path,
    )
    with app.test_request_context('/setup/finalize'):
        app.create_db()
        db_session = db.session()
        admin = User(
            email=admin_account['email'],
            password=admin_account['password'],
            last_name=admin_account['name'],
            first_name=admin_account['firstname'],
            can_login=True,
        )
        db_session.add(admin)
        security = get_service('security')
        security.grant_role(admin, Admin)
        db_session.commit()

    session_clear()

    return render_template(
        'setupwizard/done.html',
        config_file=config_file,
        logging_file=logging_file,
    )



# coding=utf-8
"""Useful decorators for web views."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

# We're now using the `deprecated` decorator from the `python-deprecated`
# package. Direct use of this package is not recommended.
from deprecated import deprecated

__all__ = ('deprecated',)



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import uuid

from pytest import raises

from . import repository, session_repository
from .service import RepositoryTransaction


def test_transaction_lifetime(session):
    state = session_repository.app_state
    root_transaction = state.get_transaction(session)
    assert isinstance(root_transaction, RepositoryTransaction)
    assert root_transaction._parent is None

    # create sub-transaction (db savepoint)
    session.begin(nested=True)
    transaction = state.get_transaction(session)
    assert isinstance(transaction, RepositoryTransaction)
    assert transaction._parent is root_transaction

    session.flush()
    transaction = state.get_transaction(session)

    # FIXME
    # assert transaction is root_transaction
    #
    # # create subtransaction (sqlalchemy)
    # session.begin(subtransactions=True)
    # transaction = state.get_transaction(session)
    # assert isinstance(transaction, RepositoryTransaction)
    # assert transaction._parent is root_transaction
    #
    # session.flush()
    # transaction = state.get_transaction(session)
    # assert transaction is root_transaction


def test_accessors_bad_uuid_type(session):
    uuid_str = b'4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9'

    with raises(ValueError):
        session_repository.get(session, uuid_str)
    with raises(ValueError):
        session_repository.set(session, uuid_str, '')
    with raises(ValueError):
        session_repository.delete(session, uuid_str)


def test_accessors_non_existent_entry(session):
    # non-existent
    u = uuid.uuid4()
    null = object()
    assert session_repository.get(session, u) is None
    assert session_repository.get(session, u, default=null) is null


def test_accessors_set_get_delete(session):
    # set
    content = b'my file content'
    u1 = uuid.uuid4()
    session_repository.set(session, u1, content)
    assert session_repository.get(session, u1).open('rb').read() == content
    assert repository.get(u1) is None

    # delete
    session_repository.delete(session, u1)
    assert session_repository.get(session, u1) is None

    u2 = uuid.uuid4()
    repository.set(u2, b'existing content')
    assert session_repository.get(session, u2) is not None

    session_repository.delete(session, u2)
    assert session_repository.get(session, u2) is None
    assert repository.get(u2) is not None


def test_transaction(session):
    u = uuid.uuid4()
    repository.set(u, b'first draft')
    assert session_repository.get(session, u).open('rb').read() == \
        b'first draft'

    session_repository.set(session, u, b'new content')

    # test nested (savepoint)
    # delete content but rollback transaction
    db_tr = session.begin(nested=True)
    session_repository.delete(session, u)
    assert session_repository.get(session, u) is None

    db_tr.rollback()
    assert session_repository.get(session, u).open('rb').read() == \
        b'new content'

    # delete and commit
    with session.begin(nested=True):
        session_repository.delete(session, u)
        assert session_repository.get(session, u) is None

    assert session_repository.get(session, u) is None
    assert repository.get(u) is not None

    session.commit()
    assert repository.get(u) is None

    # delete: now test subtransactions (sqlalchemy)
    repository.set(u, b'first draft')
    db_tr = session.begin(subtransactions=True)
    session_repository.delete(session, u)
    assert session_repository.get(session, u) is None

    db_tr.rollback()
    assert session_repository.get(session, u).open('rb').read() == \
        b'first draft'

    session.rollback()

    with session.begin(subtransactions=True):
        session_repository.delete(session, u)
        assert session_repository.get(session, u) is None

    assert session_repository.get(session, u) is None
    assert repository.get(u) is not None

    session.commit()
    assert repository.get(u) is None

    # now test 'set'
    session_repository.set(session, u, b'new content')
    session.commit()
    assert repository.get(u) is not None

    # test "set" in two nested transactions. This tests a specific code
    # branch, when a subtransaction overwrite data set in parent
    # transaction
    with session.begin(nested=True):
        session_repository.set(session, u, b'transaction 1')

        with session.begin(nested=True):
            session_repository.set(session, u, b'transaction 2')

        assert session_repository.get(session, u).open('rb').read() \
            == b'transaction 2'


def test_transaction_path(session):
    """Test RepositoryTransaction create storage only when needed."""
    u = uuid.uuid4()

    state = session_repository.app_state
    root_transaction = state.get_transaction(session)

    # assert not root_transaction.path.exists()

    with session.begin(subtransactions=True):
        transaction = state.get_transaction(session)
        assert not transaction.path.exists()

        session_repository.set(session, u, b'my file content')
        assert transaction.path.exists()

    assert root_transaction.path.exists()

    content = session_repository.get(session, u).open('rb').read()
    assert content == b'my file content'
    assert root_transaction.path.exists()



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import hashlib

from flask import Blueprint, Response, g, make_response, request
from sqlalchemy.sql.expression import func, or_
from werkzeug.exceptions import NotFound

from abilian.core.models.subjects import User
from abilian.web import url_for
from abilian.web.views import JSONModelSearch

bp = Blueprint('users', __name__, url_prefix='/users')


@bp.url_value_preprocessor
def get_user(endpoint, values):
    try:
        user_id = values.pop('user_id')
        user = User.query.get(user_id)
        if user:
            values['user'] = user
        else:
            raise NotFound()
    except KeyError:
        # this endpoint is not looking for a specific user
        pass


@bp.route('/<int:user_id>/photo')
def photo(user):
    if not user.photo:
        raise NotFound()

    data = user.photo
    self_photo = (user.id == g.user.id)

    if self_photo:
        # special case: for their own photo user has an etag, so that on change,
        # photo is immediatly reloaded from server.
        #
        # FIXME: there should be a photo_digest field on user object
        acc = hashlib.md5(data)
        etag = acc.hexdigest()

        if request.if_none_match and etag in request.if_none_match:
            return Response(status=304)

    response = make_response(data)  # type: Response
    response.content_type = 'image/jpeg'

    if not self_photo:
        response.headers.add('Cache-Control', 'public, max-age=600')
    else:
        # user always checks its own mugshot is up-to-date, in order to avoid
        # seeing old one immediatly after having uploaded of a new picture.
        response.headers.add('Cache-Control', 'private, must-revalidate')
        response.set_etag(etag)

    return response


# JSON search
class UserJsonListing(JSONModelSearch):

    Model = User
    minimum_input_length = 0

    def filter(self, query, q, **kwargs):
        if q:
            query = query.filter(
                or_(
                    func.lower(User.first_name).like(q + "%"),
                    func.lower(User.last_name).like(q + "%"),
                ),
            )
        return query

    def order_by(self, query):
        return query.order_by(
            func.lower(User.last_name),
            func.lower(User.first_name),
        )

    def get_item(self, obj):
        d = super(UserJsonListing, self).get_item(obj)
        d['email'] = obj.email
        d['can_login'] = obj.can_login
        d['photo'] = url_for('users.photo', user_id=obj.id)
        return d


bp.route('/json/')(UserJsonListing.as_view('json_list'))



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sys
from typing import List, Text

from blinker import Namespace, Signal
from flask import current_app, g
from flask_debugtoolbar.panels import DebugPanel
from six import text_type

from abilian.web.action import actions


class ActionDebugPanel(DebugPanel):
    name = 'Actions'

    user_enable = True
    has_content = True

    def nav_title(self):
        return 'Actions'

    def title(self):
        return 'Actions'

    def url(self):
        return ''

    def content(self):
        # type: () -> Text
        actions_for_template = []

        for category in actions.actions().keys():
            available_actions = actions.for_category(category)
            for action in available_actions:
                d = {
                    'category': action.category,
                    'title': action.title,
                    'class': action.__class__.__name__,
                }
                try:
                    d['endpoint'] = text_type(action.endpoint)
                except BaseException:
                    d['endpoint'] = '<Exception>'
                try:
                    d['url'] = text_type(action.url(g.action_context))
                except BaseException:
                    d['url'] = '<Exception>'
                actions_for_template.append(d)

        actions_for_template.sort(key=lambda x: (x['category'], x['title']))

        ctx = {'actions': actions_for_template}

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template(
            'debug_panels/actions_panel.html',
        )
        return template.render(ctx)


class SignalsDebugPanel(DebugPanel):
    name = 'Signals'

    user_enable = True
    has_content = True

    events = []  # type: List[dict]

    def nav_title(self):
        return 'Signals'

    def title(self):
        return 'Signals'

    def url(self):
        return ''

    def content(self):
        # type: () -> Text
        module_names = sorted(sys.modules.keys())

        signals = []

        for module_name in module_names:
            module = sys.modules[module_name]
            if not module:
                continue
            module_vars = vars(module)
            for var_name, var in module_vars.items():
                if not isinstance(var, Namespace):
                    continue

                ns = var
                ns_name = var_name
                for signal_name, signal in ns.items():
                    receivers = [
                        text_type(r) for r in signal.receivers.values()
                    ]
                    d = {
                        'module_name': module_name,
                        'ns_name': ns_name,
                        'signal_name': signal_name,
                        'signal': signal,
                        'receivers': receivers,
                    }
                    signals.append(d)

        signals.sort(
            key=lambda x: (d['module_name'], d['ns_name'], d['signal_name']),
        )

        ctx = {'signals': signals, 'events': self.events}

        jinja_env = current_app.jinja_env
        jinja_env.filters.update(self.jinja_env.filters)
        template = jinja_env.get_or_select_template(
            'debug_panels/signals_panel.html',
        )
        return template.render(ctx)

    def process_request(self, request):
        self.events = []

        if getattr(Signal.send, '__wrapped', False):
            return

        orig_send = Signal.send
        events = self.events

        def wrapped_send(self, *sender, **kwargs):
            d = {
                'signal_name': self.name,
                'sender': text_type(sender[0]),
                'args': kwargs,
            }
            events.append(d)
            return orig_send(self, *sender, **kwargs)

        wrapped_send.__wrapped = True
        Signal.send = wrapped_send



# coding=utf-8
"""Configuration and injectable fixtures for Pytest.

Supposed to replace the too-complex current UnitTest-based testing
framework.

DI and functions over complex inheritance hierarchies FTW!
"""

from __future__ import absolute_import, division, print_function, \
    unicode_literals

import os
import warnings

import pytest

pytest_plugins = ['abilian.testing.fixtures']

if os.environ.get("FAIL_ON_WARNINGS"):
    # Don't remove !
    import pandas
    warnings.simplefilter("error")

if os.environ.get("COLLECT_ANNOTATIONS"):
    import pyannotate_runtime

    def pytest_collection_finish(session):
        """Handle the pytest collection finish hook: configure pyannotate.
        Explicitly delay importing `collect_types` until all tests have
        been collected.  This gives gevent a chance to monkey patch the
        world before importing pyannotate.
        """
        from pyannotate_runtime import collect_types
        collect_types.init_types_collection()

    @pytest.fixture(autouse=True)
    def collect_types_fixture():
        from pyannotate_runtime import collect_types
        collect_types.resume()
        yield
        collect_types.pause()

    def pytest_sessionfinish(session, exitstatus):
        from pyannotate_runtime import collect_types
        collect_types.dump_stats("type_info.json")



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from abilian.core.entities import Entity
from abilian.services import security


def test_default_permissions(app, db, session):

    class MyRestrictedType(Entity):
        __default_permissions__ = {
            security.READ: {security.Anonymous},
            security.WRITE: {security.Owner},
            security.CREATE: {security.Writer},
            security.DELETE: {security.Owner},
        }

    assert isinstance(MyRestrictedType.__default_permissions__, frozenset)

    expected = frozenset({
        (security.READ, frozenset({security.Anonymous})),
        #
        (security.WRITE, frozenset({security.Owner})),
        #
        (security.CREATE, frozenset({security.Writer})),
        #
        (security.DELETE, frozenset({security.Owner})),
    })
    assert MyRestrictedType.__default_permissions__ == expected

    db.create_all()  # create missing 'mytype' table

    obj = MyRestrictedType(name='test object')
    session.add(obj)
    PA = security.PermissionAssignment
    query = session.query(PA.role) \
        .filter(PA.object == obj)

    assert query.filter(PA.permission == security.READ).all() \
        == [(security.Anonymous,)]

    assert query.filter(PA.permission == security.WRITE).all() \
        == [(security.Owner,)]

    assert query.filter(PA.permission == security.DELETE).all() \
        == [(security.Owner,)]

    # special case:
    assert query.filter(PA.permission == security.CREATE).all() \
        == []

    security_svc = app.services['security']
    permissions = security_svc.get_permissions_assignments(obj)
    assert permissions == {
        security.READ: {security.Anonymous},
        security.WRITE: {security.Owner},
        security.DELETE: {security.Owner},
    }



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
import logging
import os
import re
from functools import partial
from io import StringIO
from os.path import isabs
from pathlib import Path

from webassets.filter import ExternalTool, Filter, get_filter, register_filter
from webassets.filter.closure import ClosureJS as BaseClosureJS
from webassets.utils import working_directory


class ImportCSSFilter(Filter):
    """This filter searches (recursively) '@import' rules and replaces them by
    content of target file."""
    name = 'cssimporter'
    max_debug_level = None

    logger = logging.getLogger(__name__ + '.ImportCssFilter')
    _IMPORT_RE = re.compile(
        '''@import ("|')(?P<filename>(/?[-a-zA-Z0-9_\.]+)+\.css)("|');''',
    )

    def input(self, _in, out, **kwargs):
        filepath = kwargs['source_path']
        source = kwargs.get('source')

        if not source:
            # this happens when this filters is not used as a "source" filter, i.e _in
            # is a webasset hunk instance
            source = filepath
        # output = kwargs['output']
        base_dir = Path(filepath).parent
        rel_dir = Path(source).parent

        self.logger.debug('process "%s"', filepath)

        for line in _in.readlines():
            import_match = self._IMPORT_RE.search(line)
            if import_match is None:
                out.write(line)
                continue

            filename = import_match.group('filename')
            abs_filename = os.path.abspath(base_dir / filename)
            rel_filename = os.path.normpath(rel_dir / filename)

            start, end = import_match.span()
            if start > 0:
                out.write(line[:start])
                out.write('\n')

            with open(abs_filename, 'r') as included:
                # rewrite url() statements
                buf = StringIO()
                url_rewriter = get_filter('cssrewrite')
                url_rewriter.set_context(self.ctx)
                url_rewriter.setup()
                url_rewriter.input(
                    included,
                    buf,
                    source=rel_filename,
                    source_path=abs_filename,
                    output=source,
                    output_path=filepath,
                )
            buf.seek(0)
            # now process '@includes' directives in included file
            self.input(
                buf,
                out,
                source=rel_filename,
                source_path=abs_filename,
                output=source,
                output_path=filepath,
            )

            if end < len(line):
                out.write(line[end:])
            else:
                out.write('\n')


class LessImportFilter(Filter):
    """This filter outputs `@import` statements for listed files.

    This allows to generate a single less file for application, where
    abilian properties can be overridden by application.
    """
    name = 'less_import'
    options = {
        'run_in_debug': 'LESS_RUN_IN_DEBUG',  # use same option as less filter
    }
    max_debug_level = None

    logger = logging.getLogger(__name__ + '.LessImportFilter')

    def setup(self):
        super(LessImportFilter, self).setup()
        if self.run_in_debug is False:
            # Disable running in debug mode for this instance.
            self.max_debug_level = False

    def input(self, _in, out, source_path, output_path, **kwargs):
        if not Path(source_path).is_file():
            # we are not processing files but webassets intermediate hunks
            out.write(_in.read())
            return

        out_dir = os.path.dirname(output_path)
        rel_path = os.path.relpath(source_path, out_dir)

        # note: when import as CSS, import statement is put at the top of the
        # generated file (order of import is not preserved, less content will be
        # after pure css one). If we use "inline" the lessc will not rewrite
        # url(). So we better have all our css imported as less content.
        import_mode = 'less'  # if not rel_path.endswith('css') else 'css'
        out.write('@import ({}) "{}";'.format(import_mode, rel_path))


class Less(ExternalTool):
    """Converts `less <http://lesscss.org/>`_ markup to real CSS.

    This depends on the NodeJS implementation of less, installable via npm.
    To use the old Ruby-based version (implemented in the 1.x Ruby gem), see
    :class:`~.less_ruby.Less`.

    *Supported configuration options*:

    LESS_BIN (binary)
      Path to the less executable used to compile source files. By default,
      the filter will attempt to run ``lessc`` via the system path.

    LESS_LINE_NUMBERS (line_numbers)
      Outputs filename and line numbers. Can be either 'comments', which
      will output the debug info within comments, 'mediaquery' that will
      output the information within a fake media query which is compatible
      with the SASSPath to the less executable used to compile source files.

    LESS_RUN_IN_DEBUG (run_in_debug)
      By default, the filter will compile in debug mode. Since the less
      compiler is written in Javascript and capable of running in the
      browser, you can set this to ``False`` to have your original less
      source files served (see below).

    LESS_PATHS (paths)
      Add include paths for less command line.
      It should be a list of paths relatives to Environment.directory or absolute paths.
      Order matters as less will pick the first file found in path order.

    .. admonition:: Compiling less in the browser

      less is an interesting case because it is written in Javascript and
      capable of running in the browser. While for performance reason you
      should prebuild your stylesheets in production, while developing you
      may be interested in serving the original less files to the client,
      and have less compile them in the browser.

      To do so, you first need to make sure the less filter is not applied
      when :attr:`Environment.debug` is ``True``. You can do so via an
      option::

          env.config['less_run_in_debug'] = False

      Second, in order for the less to identify the  less source files as
      needing to be compiled, they have to be referenced with a
      ``rel="stylesheet/less"`` attribute. One way to do this is to use the
      :attr:`Bundle.extra` dictionary, which works well with the template
      tags that webassets provides for some template languages::

          less_bundle = Bundle(
              '**/*.less',
              filters='less',
              extra={'rel': 'stylesheet/less' if env.debug else 'stylesheet'}
          )

      Then, for example in a Jinja2 template, you would write::

          {% assets less_bundle %}
              <link rel="{{ EXTRA.rel }}" type="text/css" href="{{ ASSET_URL }}">
          {% endassets %}

      With this, the ``<link>`` tag will sport the correct ``rel`` value both
      in development and in production.

      Finally, you need to include the less compiler::

          if env.debug:
              js_bundle.contents += \
                 'http://lesscss.googlecode.com/files/less-1.3.0.min.js'
    """

    name = 'less'
    options = {
        'less': ('binary', 'LESS_BIN'),
        'run_in_debug': 'LESS_RUN_IN_DEBUG',
        'line_numbers': 'LESS_LINE_NUMBERS',
        'extra_args': 'less_extra_args',
        'paths': 'LESS_PATHS',
        'as_output': 'less_as_output',
        'source_map_file': 'less_source_map_file',
    }
    max_debug_level = None

    logger = logging.getLogger(__name__ + '.LessFilter')

    def setup(self):
        super(Less, self).setup()
        if self.run_in_debug is False:
            # Disable running in debug mode for this instance.
            self.max_debug_level = False

    def input(self, in_, out, **kw):
        if self.as_output:
            importer = get_filter('less_import')
            importer.input(in_, out, **kw)
        else:
            self._apply_less(in_, out, **kw)

    def output(self, in_, out, **kw):
        if not self.as_output:
            out.write(in_.read())
        else:
            self._apply_less(in_, out, **kw)

    def _apply_less(self, in_, out, output_path, output, **kw):
        # Set working directory to the source file so that includes are found
        args = [self.less or 'lessc']
        if self.line_numbers:
            args.append('--line-numbers=%s' % self.line_numbers)

        if self.paths:
            paths = [
                path if isabs(path) else self.ctx.resolver.resolve_source(path)
                for path in self.pathsep
            ]
            args.append('--include-path={}'.format(os.pathsep.join(paths)))

        #
        # Commented out since this doesn't work with the current lessc compiler.
        # (See also just below)
        #
        # source_map = self.source_map_file and self.ctx.debug
        # if source_map:
        #     source_map_dest = os.path.join(self.ctx.directory,
        #                                    self.source_map_file)
        #     self.logger.debug('Generate source map to "%s"', source_map_dest)
        #     args.append('--source-map={}'.format(source_map_dest))
        #     args.append('--source-map-url={}'.format(self.source_map_file))

        if self.extra_args:
            args.extend(self.extra_args)

        args.append('-')
        buf = StringIO()
        with working_directory(filename=output_path):
            self.subprocess(args, buf, in_)

        # if source_map:
        #     self.fix_source_map_urls(source_map_dest)

        # rewrite css url()
        replace_url = partial(self.fix_url, os.path.dirname(output_path))
        buf.seek(0)
        url_rewriter = get_filter('cssrewrite', replace=replace_url)
        url_rewriter.set_context(self.ctx)
        url_rewriter.setup()
        url_rewriter.input(
            buf,
            out,
            source=output,
            source_path=output_path,
            output=output,
            output_path=output_path,
        )

    def fix_url(self, cur_path, url):
        if url.startswith('data:'):
            # base64 embeded
            return url

        src_path = os.path.normpath(
            os.path.abspath(os.path.join(cur_path, url)),
        )
        possible_paths = [
            p for p in self.ctx.url_mapping.keys() if src_path.startswith(p)
        ]
        if not possible_paths:
            return url

        if len(possible_paths) > 1:
            raise RuntimeError("Should not happen")
            # possible_paths.sort(lambda p: -len(p))

        path = possible_paths[0]
        return self.ctx.url_mapping[path] + src_path[len(path):]

    def fix_source_map_urls(self, filename):
        with open(filename, 'r') as f:
            data = json.load(f)

        for idx, path in enumerate(data['sources']):
            if path == '-':
                data['sources'][idx] = '-'
                continue

            # apparently less is stripping first part
            path = os.path.join('..', path)
            data['sources'][idx] = self.fix_url(self.ctx.directory, path)

        with open(filename, 'w') as f:
            json.dump(data, f)


class ClosureJS(BaseClosureJS):

    def setup(self):
        super(ClosureJS, self).setup()
        self.source_files = []

    def input(self, _in, out, source_path, output_path, **kwargs):
        if not Path(source_path).is_file():
            # we are not processing files but webassets intermediate hunks
            return
        self.source_files.append(source_path)

    def output(self, _in, out, **kw):
        for source_file in self.source_files:
            self.extra_args.append('--js')
            self.extra_args.append(source_file)

        super(ClosureJS, self).output(_in, out, **kw)
        try:
            smap_idx = self.extra_args.index('--create_source_map')
            smap_path = Path(self.extra_args[smap_idx + 1])
        except (ValueError, IndexError):
            return

        if not smap_path.exists():
            return

        name = smap_path.name
        out.write('//# sourceMappingURL={}'.format(str(name)))
        self.fix_source_map_urls(str(smap_path))

    def fix_url(self, cur_path, src_path):
        possible_paths = [
            p for p in self.ctx.url_mapping.keys() if src_path.startswith(p)
        ]

        if not possible_paths:
            raise RuntimeError("Should not happen")
            # # FIXME: url is not defined at this point, this can't work.
            # return url

        if len(possible_paths) > 1:
            raise RuntimeError("Should not happen")
            # possible_paths.sort(lambda p: -len(p))

        path = possible_paths[0]
        return self.ctx.url_mapping[path] + src_path[len(path):]

    def fix_source_map_urls(self, filename):
        with open(filename, 'r') as f:
            data = json.load(f)

        for idx, path in enumerate(data['sources']):
            if path == '-':
                data['sources'][idx] = '-'
                continue

            data['sources'][idx] = self.fix_url(self.ctx.directory, path)

        with open(filename, 'w') as f:
            json.dump(data, f)


def register_filters():
    register_filter(Less)
    register_filter(LessImportFilter)
    register_filter(ImportCSSFilter)
    register_filter(ClosureJS)


register_filters()



# coding=utf-8
"""Base classes for file download."""
from __future__ import absolute_import, print_function, unicode_literals

from datetime import datetime, timedelta

from flask import Response, request, send_file
from werkzeug.exceptions import BadRequest

from abilian.core.util import utc_dt

from .base import View


class BaseFileDownload(View):

    set_expire = False
    expire_offset = timedelta(days=365)
    as_attachment = False

    #: argument name that must be found in view kwargs. This is a safety measure
    #: to prevent setting far expire date on resources without a varying argument
    #: in url (path or query string), such as a timestamp, a serial, a hash...
    expire_vary_arg = None

    def __init__(
        self,
        set_expire=None,
        expire_offset=None,
        expire_vary_arg=None,
        as_attachment=None,
    ):
        # Override class default value only if arg is specified in constructor.
        # This allows subclasses to easily override these defaults.
        if set_expire is not None:
            self.set_expire = set_expire
        if expire_offset is not None:
            self.expire_offset = expire_offset
        if expire_vary_arg is not None:
            self.expire_vary_arg = expire_vary_arg
        if as_attachment is not None:
            self.as_attachment = bool(as_attachment)

        if self.set_expire:
            if not self.expire_offset:
                raise ValueError('expire_offset is not set')
            if not self.expire_vary_arg:
                raise ValueError('expire_vary_arg is not set')

    def prepare_args(self, args, kwargs):
        if self.set_expire:
            vary_arg = kwargs.get(
                self.expire_vary_arg,
                request.args.get(self.expire_vary_arg),
            )
            if vary_arg is None:
                # Argument for timestamp, serial etc is missing.
                # We must refuse to serve an image with expiry date set up
                # to maybe 1 year from now.
                # Check the code that has generated this url!
                raise BadRequest(
                    'File version marker is missing ({}=?)'.format(
                        repr(self.expire_vary_arg),
                    )
                )

        args, kwargs = super(BaseFileDownload, self).prepare_args(args, kwargs)
        kwargs['attach'] = request.args.get(
            'attach',
            self.as_attachment,
            type=bool,
        )
        return args, kwargs

    def make_response(self, *args, **kwargs):
        # for example: return flask.make_response(...)
        # or: return flask.send_file(...)
        raise NotImplementedError()

    def get_filename(self, *args, **kwargs):
        raise NotImplementedError()

    def get_content_type(self, *args, **kwargs):
        raise NotImplementedError()

    def get(self, attach, *args, **kwargs):
        """
        :param attach: if True, return file as an attachment.
        """
        response = self.make_response(*args, **kwargs)  # type: Response
        response.content_type = self.get_content_type(*args, **kwargs)

        if attach:
            filename = self.get_filename(*args, **kwargs)
            if not filename:
                filename = 'file.bin'
            headers = response.headers
            headers.add('Content-Disposition', 'attachment', filename=filename)

        self.set_cache_headers(response)
        return response

    def set_cache_headers(self, response):
        if self.set_expire:
            response.cache_control.public = False
            response.cache_control.private = True
            response.cache_control.max_age = int(
                self.expire_offset.total_seconds(),
            )
            response.expires = utc_dt(datetime.utcnow() + self.expire_offset)


class BaseBlobDownload(BaseFileDownload):

    def get_blob(self, *args, **kwargs):
        raise NotImplementedError()

    def prepare_args(self, args, kwargs):
        args, kwargs = super(BaseBlobDownload, self).prepare_args(args, kwargs)
        self.blob = self.get_blob(*args, **kwargs)
        metadata = self.blob.meta
        self.filename = metadata.get('filename', self.obj.name)
        self.content_type = metadata.get('mimetype')
        return args, kwargs

    def get_filename(self, *args, **kwargs):
        return self.filename

    def get_content_type(self, *args, **kwargs):
        return self.content_type

    def make_response(self, *args, **kwargs):
        blob = self.blob
        stream = blob.file.open('rb')

        return send_file(
            stream,
            mimetype=self.content_type,
            cache_timeout=0,
            add_etags=False,
        )



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, unicode_literals

from flask import Blueprint, redirect
from werkzeug.datastructures import ImmutableDict

from abilian.app import Application as BaseApplication
from abilian.core.celery import FlaskCelery as BaseCelery
from abilian.core.celery import FlaskLoader as CeleryBaseLoader
from abilian.services.vocabularies import Vocabulary
from abilian.web.util import url_for

from .config import Config

__all__ = ['create_app']

APP_NAME = 'abilian_core_demo'


def create_app(config=None):
    return Application(config=config)


# loader to be used by celery workers
class CeleryLoader(CeleryBaseLoader):
    flask_app_factory = 'demo.app.create_app'


class CeleryApp(BaseCelery):
    loader_cls = CeleryLoader


celery = CeleryApp(loader=CeleryLoader)

main = Blueprint("main", __name__, url_prefix="")


@main.route("/")
def home():
    """
    Home page. Actually there is no home page, so for this demo
    we redirect to the most appropriate place.
    """
    return redirect(url_for('admin.settings'))


default_config = dict(BaseApplication.default_config)
default_config.update(dict(vars(Config)))
default_config = ImmutableDict(default_config)


class Application(BaseApplication):
    default_config = default_config
    celery_app_cls = CeleryApp

    def __init__(self, name=APP_NAME, config=None, **kwargs):
        super(Application, self).__init__(
            name, config=config, instance_relative_config=True, **kwargs
        )

    def init_extensions(self):
        super(Application, self).init_extensions()

        # Additional service
        from abilian.services.security import security
        security.init_app(self)

    def register_plugins(self):
        super(Application, self).register_plugins()
        self.register_blueprint(main)


Vocabulary_demo = Vocabulary(group='Test', name='Test', label='This is a test')



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import sqlalchemy as sa
import sqlalchemy.exc
from pytest import mark, raises

from abilian.web import url_for

from .models import BaseVocabulary, Vocabulary
from .service import vocabularies

# DefaultVoc = Vocabulary('defaultstates', group='', label='States')

# XXX: we need to use globals for now.
# Classes are global state in Python.
Voc = Vocabulary('voc', label='Voc', group='')
PriorityVoc = Vocabulary('priorities', label='Priorities', group='')
StateVoc = Vocabulary('defaultstates', group='', label='States')
DocCatVoc = Vocabulary('categories', group='documents', label='Categories')


def test_vocabulary_factory(session):
    assert Voc.__name__ == 'VocabularyVoc'
    assert Voc.__tablename__ == 'vocabulary_voc'
    assert Voc.Meta.name == 'voc'
    assert Voc.Meta.label == 'Voc'
    assert Voc.Meta.group is ""
    assert issubclass(Voc, BaseVocabulary)


def test_vocabularies(session):
    # test registered vocabularies
    assert vocabularies.vocabularies == {Voc, PriorityVoc, StateVoc, DocCatVoc}

    grouped = vocabularies.grouped_vocabularies
    assert set(grouped['']) == {Voc, StateVoc, PriorityVoc}
    assert grouped['documents'] == [DocCatVoc]

    assert vocabularies.get_vocabulary('priorities') is PriorityVoc
    assert vocabularies.get_vocabulary('priorities', 'nogroup') is None
    assert vocabularies.get_vocabulary('categories', 'documents') is DocCatVoc


def test_items(db, session):
    db.create_all()

    IMMEDIATE = PriorityVoc(label='Immediate', position=0)
    NORMAL = PriorityVoc(label='Normal', position=3, default=True)
    URGENT = PriorityVoc(label='Urgent', position=1)
    HIGH = PriorityVoc(label='High', position=2)
    items = (IMMEDIATE, NORMAL, URGENT, HIGH)
    for item in items:
        session.add(item)
    session.flush()

    # test position=4 set automatically; Label stripped
    low_item = PriorityVoc(label=' Low  ')
    session.add(low_item)
    session.flush()
    assert low_item.position == 4
    assert low_item.label == 'Low'

    # test strip label on update
    IMMEDIATE.label = '  Immediate  '
    session.flush()
    assert IMMEDIATE.label == 'Immediate'

    # test default ordering
    default_ordering = ['Immediate', 'Urgent', 'High', 'Normal', 'Low']
    query = PriorityVoc.query
    assert [i.label for i in query.active().all()] == default_ordering

    # no default ordering when using .values(): explicit ordering required
    query = PriorityVoc.query \
        .active() \
        .order_by(PriorityVoc.position.asc())
    assert ([i.label for i in query.values(PriorityVoc.label)
            ] == default_ordering)

    # test db-side constraint for non-empty labels
    with raises(
        sa.exc.IntegrityError,
        message="Could insert an item with empty label",
    ):
        with session.begin_nested():
            v = PriorityVoc(label='   ', position=6)
            session.add(v)
            session.flush()

    # test unique labels constraint
    with raises(sa.exc.IntegrityError, message="Could insert duplicate label"):
        with session.begin_nested():
            v = PriorityVoc(label='Immediate')
            session.add(v)
            session.flush()

    # test unique position constraint
    with raises(
        sa.exc.IntegrityError,
        message="Could insert duplicate position",
    ):
        with session.begin_nested():
            v = PriorityVoc(label='New one', position=1)
            session.add(v)
            session.flush()

    # test by_position without results
    item = PriorityVoc.query.by_position(42)
    assert item is None

    # test by_position() and active()
    item = PriorityVoc.query.by_position(URGENT.position)
    assert item is URGENT

    item.active = False
    expected = [
        'Immediate',
        'High',
        'Normal',
        'Low',
    ]
    assert ([i.label for i in PriorityVoc.query.active().all()] == expected)
    assert PriorityVoc.query.active().by_position(URGENT.position) is None

    # test by_label()
    item = PriorityVoc.query.by_label(NORMAL.label)
    assert item is NORMAL


@mark.skip
def test_admin_panel_reorder(app, db, session, client, test_request_context):
    db.create_all()
    items = [
        Voc(label='First', position=0),
        Voc(label='Second', position=2),
        Voc(label='Third', position=3),
    ]

    for i in items:
        session.add(i)
    session.commit()

    first, second, third = items
    url = url_for('admin.vocabularies')
    base_data = {'Model': Voc.Meta.name}
    data = {'down': first.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert r.headers['Location'] == 'http://localhost/admin/vocabularies'
    assert Voc.query.order_by(Voc.position).all() == [second, first, third]

    data = {'up': first.id, 'return_to': 'group'}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert r.headers['Location'] \
        == 'http://localhost/admin/vocabularies/_/'
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {'up': first.id, 'return_to': 'model'}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert r.headers['Location'] \
        == 'http://localhost/admin/vocabularies/_/defaultstates/'
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {'down': third.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert r.headers['Location'] == 'http://localhost/admin/vocabularies'
    assert Voc.query.order_by(Voc.position).all() == [first, second, third]

    data = {'up': third.id}
    data.update(base_data)
    r = client.post(url, data=data)
    assert r.status_code == 302
    assert r.headers['Location'] == 'http://localhost/admin/vocabularies'
    assert Voc.query.order_by(Voc.position).all() == [first, third, second]



# coding=utf-8
"""Additional data types for sqlalchemy."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
import logging
import sys
import uuid

import babel
import babel.dates
import flask_sqlalchemy as flask_sa
import pytz
import sqlalchemy as sa
import sqlalchemy.dialects
import sqlalchemy.exc
import sqlalchemy.orm
import sqlalchemy.pool
from flask_sqlalchemy import SQLAlchemy as SAExtension
from six import string_types, text_type
from sqlalchemy.event import listens_for
from sqlalchemy.ext.mutable import Mutable

from .logging import patch_logger

logger = logging.getLogger(__name__)


@listens_for(sa.pool.Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    """Ensure connections are valid.

    From: `http://docs.sqlalchemy.org/en/rel_0_8/core/pooling.html`

    In case db has been restarted pool may return invalid connections.
    """
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except BaseException:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise sa.exc.DisconnectionError()
    cursor.close()


class SQLAlchemy(SAExtension):
    """Base subclass of :class:`flask_sqlalchemy.SQLAlchemy`.

    Add our custom driver hacks.
    """

    def apply_driver_hacks(self, app, info, options):
        SAExtension.apply_driver_hacks(self, app, info, options)

        if info.drivername == 'sqlite':
            connect_args = options.setdefault('connect_args', {})

            if 'isolation_level' not in connect_args:
                # required to support savepoints/rollback without error. It disables
                # implicit BEGIN/COMMIT statements made by pysqlite (a COMMIT kills all
                # savepoints made).
                connect_args['isolation_level'] = None
        elif info.drivername.startswith('postgres'):
            options.setdefault('client_encoding', 'utf8')


# PATCH flask_sqlalchemy for proper info in debug toolbar.
#
# Original code works only when current app code is involved. If using 3rd party
# app the query is logged but source is marked "unknown". Our patch is a "best
# guess".
def _calling_context(app_path):
    frm = sys._getframe(1)
    entered_sa_code = exited_sa_code = False
    sa_caller = '<unknown>'
    format_name = (
        '{frm.f_code.co_filename}:{frm.f_lineno} '
        '({frm.f_code.co_name})'.format
    )

    while frm.f_back is not None:
        name = frm.f_globals.get('__name__')
        if name and (
            name == app_path or name.startswith(app_path + '.') or
            name.startswith('abilian.')
        ):
            return format_name(frm=frm)

        if not exited_sa_code:
            in_sa_code = name and (
                name == 'sqlalchemy' or name.startswith('sqlalchemy.')
            )
            if not entered_sa_code:
                entered_sa_code = in_sa_code
            elif not in_sa_code:
                # exited from sa stack: retain name
                sa_caller = format_name(frm=frm)
                exited_sa_code = True

        frm = frm.f_back

    return sa_caller


patch_logger.info(flask_sa._calling_context)
flask_sa._calling_context = _calling_context
del flask_sa

# END PATCH


def filter_cols(model, *filtered_columns):
    """Return columnsnames for a model except named ones.

    Useful for defer() for example to retain only columns of interest
    """
    m = sa.orm.class_mapper(model)
    return list({p.key
                 for p in m.iterate_properties
                 if hasattr(p, 'columns')}.difference(filtered_columns),)


class MutationDict(Mutable, dict):
    """Provides a dictionary type with mutability support."""

    @classmethod
    def coerce(cls, key, value):
        """Convert plain dictionaries to MutationDict."""
        if not isinstance(value, MutationDict):
            if isinstance(value, dict):
                return MutationDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    #  pickling support. see:
    #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
    def __getstate__(self):
        return dict(self)

    def __setstate__(self, state):
        self.update(state)

    # dict methods
    def __setitem__(self, key, value):
        """Detect dictionary set events and emit change events."""
        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        """Detect dictionary del events and emit change events."""
        dict.__delitem__(self, key)
        self.changed()

    def clear(self):
        dict.clear(self)
        self.changed()

    def update(self, other):
        dict.update(self, other)
        self.changed()

    def setdefault(self, key, failobj=None):
        if key not in self:
            self.changed()
        return dict.setdefault(self, key, failobj)

    def pop(self, key, *args):
        self.changed()
        return dict.pop(self, key, *args)

    def popitem(self):
        self.changed()
        return dict.popitem(self)


class MutationList(Mutable, list):
    """Provides a list type with mutability support."""

    @classmethod
    def coerce(cls, key, value):
        """Convert list to MutationList."""
        if not isinstance(value, MutationList):
            if isinstance(value, list):
                return MutationList(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    #  pickling support. see:
    #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_parents', None)
        return d

    # list methods
    def __setitem__(self, idx, value):
        list.__setitem__(self, idx, value)
        self.changed()

    def __delitem__(self, idx):
        list.__delitem__(self, idx)
        self.changed()

    def insert(self, idx, value):
        list.insert(self, idx, value)
        self.changed()

    # def __setslice__(self, i, j, other):
    #     list.__setslice__(self, i, j, other)
    #     self.changed()
    #
    # def __delslice__(self, i, j):
    #     list.__delslice__(self, i, j)
    #     self.changed()

    def __iadd__(self, other):
        l = list.__iadd__(self, other)
        self.changed()
        return l

    def __imul__(self, n):
        l = list.__imul__(self, n)
        self.changed()
        return l

    def append(self, item):
        list.append(self, item)
        self.changed()

    def pop(self, i=-1):
        item = list.pop(self, i)
        self.changed()
        return item

    def remove(self, item):
        list.remove(self, item)
        self.changed()

    def reverse(self):
        list.reverse(self)
        self.changed()

    def sort(self, *args, **kwargs):
        list.sort(self, *args, **kwargs)
        self.changed()

    def extend(self, other):
        list.extend(self, other)
        self.changed()


class JSON(sa.types.TypeDecorator):
    """Stores any structure serializable with json.

    Usage JSON() Takes same parameters as sqlalchemy.types.Text
    """
    impl = sa.types.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class JSONUniqueListType(JSON):
    """Store a list in JSON format, with items made unique and sorted."""

    @property
    def python_type(self):
        return MutationList

    def process_bind_param(self, value, dialect):
        # value may be a simple string used in a LIKE clause for instance, so we
        # must ensure we uniquify/sort only for list-like values
        if value is not None and isinstance(value, (tuple, list)):
            value = sorted(set(value))

        return JSON.process_bind_param(self, value, dialect)


def JSONDict(*args, **kwargs):
    """Stores a dict as JSON on database, with mutability support."""
    return MutationDict.as_mutable(JSON(*args, **kwargs))


def JSONList(*args, **kwargs):
    """Stores a list as JSON on database, with mutability support.

    If kwargs has a param `unique_sorted` (which evaluated to True),
    list values are made unique and sorted.
    """
    type_ = JSON
    try:
        if kwargs.pop('unique_sorted'):
            type_ = JSONUniqueListType
    except KeyError:
        pass

    return MutationList.as_mutable(type_(*args, **kwargs))


class UUID(sa.types.TypeDecorator):
    """Platform-independent UUID type.

    Uses Postgresql's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    From SQLAlchemy documentation.
    """
    impl = sa.types.CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(sa.dialects.postgresql.UUID())
        else:
            return dialect.type_descriptor(sa.types.CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            # hexstring
            return value.hex

    def process_result_value(self, value, dialect):
        return value if value is None else uuid.UUID(value)

    def compare_against_backend(self, dialect, conn_type):
        if dialect.name == 'postgresql':
            return isinstance(conn_type, sa.dialects.postgresql.UUID)
        else:
            return isinstance(conn_type, sa.types.CHAR)


class Locale(sa.types.TypeDecorator):
    """Store a :class:`babel.Locale` instance."""
    impl = sa.types.UnicodeText

    @property
    def python_type(self):
        return babel.Locale

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        if not isinstance(value, babel.Locale):
            if not isinstance(value, string_types):
                raise ValueError(
                    "Unknown locale value: " + format(repr(value),),
                )
            if not value.strip():
                return None
            value = babel.Locale.parse(value)

        code = text_type(value.language)
        if value.territory:
            code += '_' + text_type(value.territory)
        elif value.script:
            code += '_' + text_type(value.territory)

        return code

    def process_result_value(self, value, dialect):
        return None if value is None else babel.Locale.parse(value)


class Timezone(sa.types.TypeDecorator):
    """Store a :class:`pytz.tzfile.DstTzInfo` instance."""
    impl = sa.types.UnicodeText

    @property
    def python_type(self):
        return pytz.tzfile.DstTzInfo

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        if not isinstance(value, pytz.tzfile.DstTzInfo):
            if not isinstance(value, string_types):
                raise ValueError("Unknown timezone value: " + repr(value))
            if not value.strip():
                return None
            value = babel.dates.get_timezone(value)

        return value.zone

    def process_result_value(self, value, dialect):
        return None if value is None else babel.dates.get_timezone(value)


# SQLAlchemy > 0.9 has a function generator that forget to set __module__
# attributes in 0.9.x series. This is fixed in 1.0.x. Sphinx will try to include
# those symbols in documentation, and this may break since we don't have
# sphinx's extensions used by sqlalchemy author.
#
# Ref:
# https://bitbucket.org/zzzeek/sqlalchemy/issues/3218/__module__-should-be-set-on-functions
if not sa.orm.relationship.__module__:
    sa.orm.relationship.__module__ = 'sqlalchemy.orm'



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from six import text_type

from abilian.core.models.subjects import User


def test_non_ascii_password():
    """Ensure we can store and test non-ascii password without any
    UnicodeEncodeError."""
    user = User()

    user.set_password('Hé')

    if not isinstance(user.password, text_type):
        # when actually retrieved from database, it should be Unicode
        user.password = text_type(user.password)

    assert user.authenticate('Hé')



# coding=utf-8
""""""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from datetime import datetime

import six
import sqlalchemy as sa
from whoosh.fields import NUMERIC, TEXT, Schema

from abilian.core.entities import Entity
from abilian.core.extensions import db
from abilian.core.models.base import SEARCHABLE, IdMixin
from abilian.core.models.base import Indexable as CoreIndexable
from abilian.services.indexing.adapter import SAAdapter


class SANotAdaptable(object):
    pass


class SANotIndexable(IdMixin, db.Model):
    __tablename__ = 'sa_not_indexable'
    __indexable__ = False


class Indexable(IdMixin, CoreIndexable, db.Model):
    __tablename__ = 'sa_indexable'
    __indexation_args__ = dict(
        index_to=(
            ('related.name', ('name', 'text')),
            ('related.description', 'text'),
        ),
    )

    num = sa.Column(
        sa.Integer,
        info=SEARCHABLE | dict(index_to=(('num', NUMERIC(numtype=int)),)),
    )


class SubclassEntityIndexable(Entity):
    pass


def test_can_adapt():
    assert not SAAdapter.can_adapt(SANotAdaptable)
    assert SAAdapter.can_adapt(SANotIndexable)
    assert SAAdapter.can_adapt(Indexable)
    assert SAAdapter.can_adapt(Entity)


def test_build_attrs():
    schema = Schema()
    adapter = SAAdapter(SANotIndexable, schema)
    assert not adapter.indexable
    assert adapter.doc_attrs == {}

    adapter = SAAdapter(Entity, schema)
    assert adapter.indexable == False

    adapter = SAAdapter(SubclassEntityIndexable, schema)
    assert adapter.indexable
    assert set(adapter.doc_attrs) == {
        'object_key',
        'id',
        'name',
        'slug',
        'object_type',
        'text',
        'created_at',
        'updated_at',
        'name_prefix',
        'owner',
        'owner_name',
        'creator_name',
        'creator',
        'allowed_roles_and_users',
        'tag_ids',
        'tag_text',
    }
    assert all(lambda f: callable(f) for f in six.itervalues(adapter.doc_attrs))

    assert set(schema.names()) == {
        'object_key',
        'id',
        'object_type',
        'name',
        'slug',
        'text',
        'created_at',
        'updated_at',
        'name_prefix',
        'owner',
        'owner_name',
        'creator_name',
        'creator',
        'allowed_roles_and_users',
        'tag_ids',
        'tag_text',
    }

    schema = Schema(
        id=NUMERIC(
            numtype=int,
            bits=64,
            signed=False,
            stored=True,
            unique=True,
        ),
    )
    adapter = SAAdapter(Indexable, schema)
    assert adapter.indexable
    assert set(adapter.doc_attrs) == {'id', 'text', 'num', 'name'}
    assert all(lambda f: callable(f) for f in six.itervalues(adapter.doc_attrs))

    assert set(schema.names()) == {'id', 'text', 'num', 'name'}
    assert isinstance(schema['text'], TEXT)
    assert isinstance(schema['num'], NUMERIC)


def test_get_document(app, db):
    schema = Schema()
    adapter = SAAdapter(SubclassEntityIndexable, schema)
    expected = {
        'id': 2,
        'name': 'entity name',
        'created_at': datetime(2013, 11, 28, 16, 17, 0),
        'updated_at': datetime(2013, 11, 29, 12, 17, 58),
    }
    obj = SubclassEntityIndexable(**expected)
    obj.slug = 'entity-name'

    expected['object_type'] = 'test_adapter.SubclassEntityIndexable'
    expected['object_key'] = 'test_adapter.SubclassEntityIndexable:2'
    expected['text'] = 'entity name'
    expected['slug'] = 'entity-name'
    expected['name_prefix'] = 'entity name'
    expected['allowed_roles_and_users'] = 'role:admin'

    with app.test_request_context():
        assert adapter.get_document(obj) == expected


def test_get_document_with_schema():
    # test retrieve related attributes
    schema = Schema(
        id=NUMERIC(
            numtype=int,
            bits=64,
            signed=False,
            stored=True,
            unique=True,
        ),
    )
    adapter = SAAdapter(Indexable, schema)
    expected = dict(id=1, num=42)
    obj = Indexable(**expected)
    obj.related = type(str('Related'), (object,), dict(name=None))()
    expected['name'] = obj.related.name = 'related name'
    obj.related.description = 'description text'
    expected['text'] = obj.related.name + ' ' + obj.related.description
    doc = adapter.get_document(obj)

    assert set(doc) == {'id', 'name', 'num', 'text'}
    assert doc['id'] == 1
    assert doc['num'] == 42
    assert doc['name'] == 'related name'
    assert 'related name' in doc['text']
    assert 'description text' in doc['text']
